/*
-----------------------------------------------------------------------------
This source file is part of the GameTools Project
http://www.gametools.org

Author: Martin Szydlowski
-----------------------------------------------------------------------------
*/

#ifndef _OgreKdTree_H__
#define _OgreKdTree_H__

#include <OgreAxisAlignedBox.h>
#include <OgreWireBoundingBox.h>
#include <OgrePlane.h>
#include <OgreVector3.h>

#include <OgreRoot.h>

#include <stack>

#include "OgreKdTreeHierarchyInterface.h"

namespace Ogre
{
	class KdRenderable;
	struct SplitInfo;

	class PlaneEvent
	{
	public:
		enum Type
		{
			PET_END,
			PET_ON,
			PET_START
		};

		enum Dimension
		{
			PED_X,
			PED_Y,
			PED_Z
		};

		enum Side
		{
			PES_LEFT = 0x01,
			PES_RIGHT = 0x02,
			PES_BOTH = PES_LEFT | PES_RIGHT
		};

		PlaneEvent(): mRenderable(0), mPosition(Vector3()), mDimension(PED_X), mType(PET_ON)
		{ };

		PlaneEvent(KdRenderable *rend, const Vector3& pos, PlaneEvent::Dimension dim, PlaneEvent::Type type):
			mRenderable(rend), mPosition(pos), mDimension(dim), mType(type) 
		{ };

		~PlaneEvent() {};

		// the less operator for plane events
		// first sort by position, then by dimension and finally by type
		bool operator < (const PlaneEvent& e) const
		{
			if(mPosition[mDimension] < e.mPosition[e.mDimension])
			{
				return true;
			}

			if(mPosition[mDimension] == e.mPosition[e.mDimension])
			{
				if (mDimension < e.mDimension)
				{
					return true;
				}
				if (mDimension == e.mDimension)
				{
					if (mType < e.mType)
					{
						return true;
					}
				}
			}

			return false;
		};

		// the equals operator for tree events
		bool operator == (const PlaneEvent& e) const
		{
			return	(mPosition[mDimension] == e.mPosition[e.mDimension]) &&
				(mDimension == e.mDimension) &&
				(mType == e.mType);
		};

		bool equalsType(const PlaneEvent& e, PlaneEvent::Type t)
		{
			return	(mPosition[mDimension] == e.mPosition[e.mDimension]) &&
				(mDimension == e.mDimension) &&
				(mType == t);
		};

		void classify(const PlaneEvent& e, PlaneEvent::Side side);

		PlaneEvent clip(AxisAlignedBox& box, PlaneEvent::Dimension dim);

		Plane * getSplitPlane() const
		{
			Vector3 normal(0,0,0);
			normal[mDimension] = 1;
			return new Plane(normal, mPosition);
		}

		KdRenderable * getRenderable() const //??
		{
			return mRenderable;
		};

		PlaneEvent::Dimension getDimension() const //??
		{
			return mDimension;
		};

		// DEBUG
		String print();
	protected:
		// event info
		KdRenderable *			mRenderable;
		Vector3					mPosition;
		PlaneEvent::Dimension	mDimension;
		PlaneEvent::Type		mType;

		// ------------------------------------------------------------------------------//
		// functions to determine the cost of splitting the node parent with the plane
		// represented by this event
		// TODO discuss if these functions really belong here, OO & SE - wise
		// pros: convenient, don't have to pass internal data to the outside
		// cons: SplitInfo ... 
	public:
		// compute "global" surface area heuristic (SAH) for the plane represented by this event
		// use only with priority queue build method
		void pqSAH(Real globalSA, Real parentSA, int nLeft, int nPlane, int nRight, AxisAlignedBox& parentBox, SplitInfo& split);
		static Real pqSplitCost(Real p, Real pl, Real pr, int nLeft, int nRight, Real mu);

		// compute the surface area heuristic (SAH) for the plane represented by this event
		void SAH(const AxisAlignedBox& parent, int nLeft, int nPlane, int nRight, SplitInfo& split);
		// the variables determining the cost of a branch traversal (KT) and a leaf intersection (KI)
		static Real KT;
		static Real KI;
		// helper functions
		static Real splitCost(Real pl, Real pr, int nLeft, int nRight);
		static Real splitCost(Real pl, Real pr, int nLeft, int nRight, Real mu);
		static Real surfaceArea(const AxisAlignedBox& box);
		static Real lookupPenalty(Real p);
	protected:
		Real splitBox(const AxisAlignedBox& parent, AxisAlignedBox& left, AxisAlignedBox& right);
	};

	// Holds all the important information on a split
	struct SplitInfo
	{
		AxisAlignedBox bleft;
		AxisAlignedBox bright;
		int nleft;
		int nright;
		Real cost;
		PlaneEvent::Side side;
		PlaneEvent event;

		// DEBUG
		String print();
	};

	typedef std::list<PlaneEvent> PlaneEventList;
	typedef std::list<KdRenderable *> KdRenderableList;
	
	class KdTree
	{
	protected:
		class Branch;
		class Leaf;

		class Node
		{
		public:
			Node(int level, AxisAlignedBox& aabb, Branch * parent):
			mLevel(level),
			mAABB(aabb),
			mParent(parent),
			mWBB(0)
			{ };

			virtual ~Node()
			{
				delete mWBB;
			};

			virtual bool isLeaf() const = 0;
			virtual bool isEmpty() const = 0;
			
			// consider using typesafe callback functions
			//virtual void cbInsert(KdTree * caller, KdRenderable * rend) = 0;

			WireBoundingBox * getWireBoundingBox()
			{
				if (mWBB == 0)
					mWBB = new WireBoundingBox();

				mWBB->setupBoundingBox(mAABB);
				
#ifdef KDTREE_DEBUG
				int level = -1;
				bool boxes = false;
				if (Root::getSingleton()._getCurrentSceneManager()->getOption("HighlightLevel",&level))
				{
					if (mLevel == level)
					{
						//mWBB->setMaterial("BaseGreenNoLighting");
						mWBB->setMaterial("aabbHiLite");
						return mWBB;
					}
					else
					{
						mWBB->setMaterial("BaseWhiteNoLighting");
						if (Root::getSingleton()._getCurrentSceneManager()->getOption("ShowAllBoxes", &boxes))
						{
							if (boxes)
								return mWBB;
							else
								return 0;
						}
					}
				}
#else
				mWBB->setMaterial("BaseWhiteNoLighting");
#endif // KDTREE_DEBUG
				return mWBB;

			}

			Branch * mParent;
			int mLevel;
			AxisAlignedBox mAABB;
		protected:
			WireBoundingBox * mWBB;
		};

		class Branch : public Node
		{
		public:
			Branch(int level, AxisAlignedBox& aabb, Branch * parent, Plane * splitplane, PlaneEvent::Side side):
				Node(level, aabb, parent), 
				mSplitPlane(splitplane), 
				mLeft(0), 
				mRight(0),
				mPlaneSide(side)
			{ };

			virtual ~Branch()
			{
				delete mLeft;
				delete mRight;
				delete mSplitPlane;
			};

			virtual bool isLeaf() const { return false; };

			virtual bool isEmpty() const { return (mLeft == 0 && mRight == 0); }
			
			Node * mLeft;
			Node * mRight;
			Plane  * mSplitPlane;
			PlaneEvent::Side mPlaneSide;
		};

		class Leaf : public Node
		{
		public:
			Leaf(int level, AxisAlignedBox& aabb, Branch * parent):
			  Node(level, aabb, parent) 
			{};

			virtual ~Leaf();

			virtual bool isLeaf() const { return true; };

			virtual bool isEmpty() const { return mKdRenderables.empty(); };

			virtual void remove(KdRenderable * rend)
			{
				mKdRenderables.remove(rend);
				//mKdRenderables.erase(find(mKdRenderables.begin(), mKdRenderables.end(), rend));
			};

			virtual void insert(KdRenderable * rend)
			{
				mKdRenderables.push_back(rend);
			};

			KdRenderableList mKdRenderables;
		};

		struct SplitCandidate
		{
			SplitCandidate(PlaneEventList * e, int n, AxisAlignedBox& a, KdTree::Branch * p, Real c, Real d, SplitInfo * b, PlaneEvent::Side s):
				events(e), nObjects(n), aabb(a), parent(p), cost(c), decrease(d), best(b), side(s) { };

			bool operator < (const SplitCandidate& rhs) const
			{
				return decrease < rhs.decrease;
			};

			void operator = (const SplitCandidate& rhs)
			{
				decrease = rhs.decrease;
				cost = rhs.cost;
				nObjects = rhs.nObjects;
				side = rhs.side;
				aabb = rhs.aabb;
				events = rhs.events;
				parent = rhs.parent;
				best = rhs.best;
			};

			// DEBUG
			String print();

			Real decrease;
			Real cost;
			int nObjects;
			PlaneEvent::Side side;
			AxisAlignedBox aabb;
			PlaneEventList * events;
			KdTree::Branch * parent;
			SplitInfo * best;
		};

		//typedef std::stack<SplitCandidate> SplitCandidatePQ;
		typedef std::priority_queue<SplitCandidate> SplitCandidatePQ;

	public:
		friend class Ogre::KdTreeHierarchyInterface;

		typedef KdTree::Leaf * LeafPtr;
		typedef std::set<LeafPtr> LeafSet;

		enum BuildMethod
		{
			KDBM_RECURSIVE,
			KDBM_PRIORITYQUEUE
		};
		
		KdTree(int maxdepth);
		virtual ~KdTree();

		// DEBUG
		void KdTree::dump();
		void KdTree::calcCost();

		// insert a new scene node into an existing kd-tree
		void insert(KdRenderable * rend);
		// remove a scene node from the tree
		void remove(KdRenderable * rend);
		// function to initialize a kd-tree based on the contents of the scene
		void build(KdRenderable * sceneRoot);

		// test visibility of objects and add to render	queue
		void queueVisibleObjects(Camera* cam, RenderQueue* queue, bool onlyShadowCasters, bool showBoxes = false);

		// self-explanatory ...
		int getMaxDepth() { return mMaxDepth; };
		AxisAlignedBox getBox() { if (mKdRoot) return mKdRoot->mAABB; else return AxisAlignedBox(); };
		void setBuildMethod(BuildMethod bm) { mBuildMethod = bm; }
	protected:
		// recursive insert funciton
		void recInsert(KdTree::Node * node, KdRenderable * rend);
		// helper functions for insert
		void recInsertNew(KdTree::Branch * parent, PlaneEvent::Side side, KdRenderable * rend);
		void splitBox(const KdTree::Branch& parent, AxisAlignedBox& left, AxisAlignedBox& right);
		void rebuildSubtree(KdTree::Node * node, KdRenderable * rend);
		// build scene from a list of nodes rather than a hierarchy
		KdTree::Node * buildFromList(KdRenderableList& nodelist, KdTree::Branch * parent, AxisAlignedBox& aabb);
		void addRendToList(KdTree::Node * node, KdRenderableList& nodelist);

		// recursively delete empty nodes
		void recDelete(KdTree::Node * node);

		// find the best plane for node division
		SplitInfo * pqFindPlane(PlaneEventList * events, int nObjects, AxisAlignedBox& aabb, Real globalSA);

		// priority queue based build function
		KdTree::Node * pqBuild(PlaneEventList& events, int nObjects, AxisAlignedBox& aabb, KdTree::Branch * parent);
		// recursive build function
		KdTree::Node * recBuild(PlaneEventList& events, int nObjects, AxisAlignedBox& aabb, KdTree::Branch * parent);

		// recursive rendering function
		void recQueueVisibleObjects(KdTree::Node * node, unsigned long currentFrame, Camera* cam, RenderQueue* queue, bool onlyShadowCasters, bool showBoxes);

		// the root node of the kdtree
		KdTree::Node * mKdRoot;
		// Maximum depth of the tree
		int mMaxDepth;

		// how to build the tree
		BuildMethod mBuildMethod;

		// logfile for tree creation
		Log * mBuildLog;

		// DEBUG
		void KdTree::dump(KdTree::Node * node);
		Real KdTree::calcCost(KdTree::Node * node, Real vs);
	}; // class KdTree

} // namespace Ogre

#endif // _OgreKdTree_H__